/* TroiaPreview
 * version 0.4 date 2004-09-20
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <memory.h>
#include <string.h>

#include "Font.h"
#include "Log.h"
#include "Textures.h"
#include "Console.h"

//private constants
#define ConsoleTexSizeX 256 //size of the console-texture
#define ConsoleTexSizeY 256
#define ConsoleColorBorder 2
#define ConsoleSpaceBorder 4
#define ConsoleBorder (ConsoleColorBorder + ConsoleSpaceBorder)
#define ConsoleInnerSizeX (ConsoleTexSizeX - 2 * ConsoleBorder)
#define ConsoleInnerSizeY (ConsoleTexSizeY - 2 * ConsoleBorder)
#define ConsoleTexLineSize ((ConsoleTexSizeX * 3 + 3) & ~3)
#define ConsolePixelComponent( Y, X, I ) ConsoleData[(Y) * ConsoleTexLineSize + (X) * 3 + (I)]
#define ConsoleBackR 0x00
#define ConsoleBackG 0x00
#define ConsoleBackB 0x00
#define ConsoleBorderR 0xC0
#define ConsoleBorderG 0xC0
#define ConsoleBorderB 0xC0
#define ConsoleTextR 0xFF
#define ConsoleTextG 0xFF
#define ConsoleTextB 0x00
#define ConsoleCursorR 0xFF
#define ConsoleCursorG 0x00
#define ConsoleCursorB 0xFF
#define ConsoleTextShowSizeX (long)(ConsoleInnerSizeX / FontSizeX)
#define ConsoleTextShowSizeY (long)(ConsoleInnerSizeY / FontSizeY)
#define ConsoleLeftRightStep 15
#define ConsoleUpDownStep 3
#define ConsolePageStep (ConsoleTextShowSizeY - 2)

//private global variables
int ConsoleChanged = 1;
unsigned char ConsoleData[ConsoleTexSizeY * ConsoleTexLineSize];
char ConsoleTextShow[ConsoleTextShowSizeY][ConsoleTextShowSizeX];
long ConsoleCursorShowX, ConsoleCursorShowY;
long ConsoleTopLeftX, ConsoleTopLeftY;
char ConsoleText[ConsoleTextSizeY][ConsoleTextSizeX];
long ConsoleCursorX, ConsoleCursorY;

//check (and correct) position in console
static void ConsoleCheckPos( )
{
  //correct cursor-position
  if( ConsoleCursorX < 0 )
    ConsoleCursorX = 0;
  if( ConsoleCursorX > ConsoleTextSizeX - 1 )
    ConsoleCursorX = ConsoleTextSizeX - 1;
  if( ConsoleCursorY < 0 )
    ConsoleCursorY = 0;
  if( ConsoleCursorY > ConsoleTextSizeY - 1 )
    ConsoleCursorY = ConsoleTextSizeY - 1;

  //correct top-left-corner-position
  if( ConsoleTopLeftX < 0 )
    ConsoleTopLeftX = 0;
  if( ConsoleTopLeftX > ConsoleTextSizeX - ConsoleTextShowSizeX )
    ConsoleTopLeftX = ConsoleTextSizeX - ConsoleTextShowSizeX;
  if( ConsoleTopLeftY < 0 )
    ConsoleTopLeftY = 0;
  if( ConsoleTopLeftY > ConsoleTextSizeY - ConsoleTextShowSizeY )
    ConsoleTopLeftY = ConsoleTextSizeY - ConsoleTextShowSizeY;
}

//check (and correct) position in console and adjust so that cursor is visible
static void ConsoleCheckPosCursorVisible( )
{
  //correct cursor-position
  if( ConsoleCursorX < 0 )
    ConsoleCursorX = 0;
  if( ConsoleCursorX > ConsoleTextSizeX - 1 )
    ConsoleCursorX = ConsoleTextSizeX - 1;
  if( ConsoleCursorY < 0 )
    ConsoleCursorY = 0;
  if( ConsoleCursorY > ConsoleTextSizeY - 1 )
    ConsoleCursorY = ConsoleTextSizeY - 1;

  //adjust top-left-corner-postition
  if( ConsoleTopLeftX > ConsoleCursorX - ConsoleTextShowSizeX + 10 )
    ConsoleTopLeftX = ConsoleCursorX - ConsoleTextShowSizeX + 10;
  if( ConsoleTopLeftX < ConsoleCursorX - ConsoleTextShowSizeX + 1 )
    ConsoleTopLeftX = ConsoleCursorX - ConsoleTextShowSizeX + 1;
  if( ConsoleTopLeftY > ConsoleCursorY - 2 )
    ConsoleTopLeftY = ConsoleCursorY - 2;
  if( ConsoleTopLeftY < ConsoleCursorY - ConsoleTextShowSizeY + 1 )
    ConsoleTopLeftY = ConsoleCursorY - ConsoleTextShowSizeY + 1;

  //correct top-left-corner-position
  if( ConsoleTopLeftX < 0 )
    ConsoleTopLeftX = 0;
  if( ConsoleTopLeftX > ConsoleTextSizeX - ConsoleTextShowSizeX )
    ConsoleTopLeftX = ConsoleTextSizeX - ConsoleTextShowSizeX;
  if( ConsoleTopLeftY < 0 )
    ConsoleTopLeftY = 0;
  if( ConsoleTopLeftY > ConsoleTextSizeY - ConsoleTextShowSizeY )
    ConsoleTopLeftY = ConsoleTextSizeY - ConsoleTextShowSizeY;
}

//print a character to the console-buffer
static int ConsolePut( char Chr )
{
  //check (and correct) position in console
  ConsoleCheckPos( );

  //no space in line
  if( ConsoleCursorX >= ConsoleTextSizeX - 2 )
    return 0;

  //put character into text-buffer
  ConsoleText[ConsoleCursorY][ConsoleCursorX] = Chr;

  //go to next position in line
  ConsoleCursorX++;

  //check (and correct) position in console and adjust so that cursor is visible
  ConsoleCheckPosCursorVisible( );

  //console has changed
  ConsoleChanged = 1;

  return 1;
}

//remove a character from the console-buffer
static int ConsoleBack( )
{
  //check (and correct) position in console
  ConsoleCheckPos( );

  //no character in line
  if( ConsoleCursorX <= 0 )
    return 0;

  //go to previos position
  ConsoleCursorX--;

  //remove character from text-buffer
  ConsoleText[ConsoleCursorY][ConsoleCursorX] = 0;

  //check (and correct) position in console and adjust so that cursor is visible
  ConsoleCheckPosCursorVisible( );

  //console has changed
  ConsoleChanged = 1;

  return 1;
}

//go to next line in console-buffer
static void ConsoleNewline( )
{
  //local variables
  long X, Y;

  //check (and correct) position in console
  ConsoleCheckPos( );

  //log this line
  Log( "Console", ConsoleText[ConsoleCursorY] );

  //last line
  if( ConsoleCursorY == ConsoleTextSizeY - 1 )
  {
    //scroll up
    ConsoleCursorY--;
    for( Y = 0; Y < ConsoleTextSizeY - 1; Y++ )
    for( X = 0; X < ConsoleTextSizeX; X++ )
      ConsoleText[Y][X] = ConsoleText[Y + 1][X];
    for( X = 0; X < ConsoleTextSizeX; X++ )
      ConsoleText[ConsoleTextSizeY - 1][X] = 0;
  }

  //go to start of next line
  ConsoleCursorX = 0;
  ConsoleCursorY++;

  //check (and correct) position in console and adjust so that cursor is visible
  ConsoleCheckPosCursorVisible( );

  //console has changed
  ConsoleChanged = 1;
}

//inistialize console
void ConsoleInit( )
{
  //local variables
  long X, Y, FntX, FntY, TexX, TexY;

  //clear console-data
  for( Y = 0; Y < ConsoleTexSizeY; Y++ )
  for( X = 0; X < ConsoleTexSizeX; X++ )
  {
    if( X >= ConsoleColorBorder && X < ConsoleTexSizeX - ConsoleColorBorder &&
      Y >= ConsoleColorBorder && Y < ConsoleTexSizeY - ConsoleColorBorder )
    {
      ConsolePixelComponent( Y, X, 0 ) = ConsoleBackB;
      ConsolePixelComponent( Y, X, 1 ) = ConsoleBackG;
      ConsolePixelComponent( Y, X, 2 ) = ConsoleBackR;
    }
    else
    {
      ConsolePixelComponent( Y, X, 0 ) = ConsoleBorderB;
      ConsolePixelComponent( Y, X, 1 ) = ConsoleBorderG;
      ConsolePixelComponent( Y, X, 2 ) = ConsoleBorderR;
    }
  }

  //clear text-show-buffer
  for( Y = 0; Y < ConsoleTextShowSizeY; Y++ )
  for( X = 0; X < ConsoleTextShowSizeX; X++ )
    ConsoleTextShow[Y][X] = 0;

  //clear text-buffer
  for( Y = 0; Y < ConsoleTextSizeY; Y++ )
  for( X = 0; X < ConsoleTextSizeX; X++ )
    ConsoleText[Y][X] = 0;

  //cursor-coordinates
  ConsoleCursorShowX = 0;
  ConsoleCursorShowY = 0;
  ConsoleTopLeftX = 0;
  ConsoleTopLeftY = 0;
  ConsoleCursorX = 0;
  ConsoleCursorY = 0;

  //put cursor into console-buffer
  for( FntY = 0; FntY < FontSizeY; FntY++ )
  for( FntX = 0; FntX < FontSizeX; FntX++ )
  {
    TexX = ConsoleCursorX * FontSizeX + FntX + ConsoleBorder;
    TexY = ConsoleInnerSizeY - (ConsoleCursorY * FontSizeY + FntY) + ConsoleBorder;
    if( (CursorData[FntY] & ((1 << (FontSizeX - 1)) >> FntX)) != 0 )
    {
      ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleCursorB;
      ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleCursorG;
      ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleCursorR;
    }
    else
    {
      ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleBackB;
      ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleBackG;
      ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleBackR;
    }
  }

  //console has changed
  ConsoleChanged = 1;
}

//print an interpreted character to the console-buffer
void ConsolePutc( char Chr )
{
  //backspace
  if( Chr == '\b' )
    ConsoleBack( );
  //newline
  if( Chr == '\n' )
    ConsoleNewline( );
  //normal character
  if( (unsigned char)Chr >= ' ' )
    ConsolePut( Chr );
}

//print text to the console-buffer
void ConsolePrint( char * Text )
{
  //print text
  for( ; *Text != 0; Text++ )
    ConsolePutc( *Text );
}

//insert a line in the console-buffer (just above the actual line)
void ConsoleInsertLine( char * Text )
{
  //local variables
  long X, Y;

  //check (and correct) position in console
  ConsoleCheckPos( );

  //log this line
  Log( "Console", Text );

  //last line
  if( ConsoleCursorY == ConsoleTextSizeY - 1 )
  {
    //scroll up
    ConsoleCursorY--;
    for( Y = 0; Y < ConsoleTextSizeY - 1; Y++ )
    for( X = 0; X < ConsoleTextSizeX; X++ )
      ConsoleText[Y][X] = ConsoleText[Y + 1][X];
    for( X = 0; X < ConsoleTextSizeX; X++ )
      ConsoleText[ConsoleTextSizeY - 1][X] = 0;
  }

  //copy current line into next line
  for( X = 0; X < ConsoleTextSizeX; X++ )
    ConsoleText[ConsoleCursorY + 1][X] = ConsoleText[ConsoleCursorY][X];

  //this instruction prevents the microsoft c compiler in visual studio 6
  //to mix up some registers when compiling in release mode ?!?!?
  ConsoleText[ConsoleCursorY][0] += 0;

  //put text into current line
  for( X = 0; Text[X] != 0 && X < ConsoleTextSizeX; X++ )
    ConsoleText[ConsoleCursorY][X] = Text[X];
  //clear rest of line
  for( ; X < ConsoleTextSizeX; X++ )
    ConsoleText[ConsoleCursorY][X] = 0;

  //go to next line but stay in current column
  ConsoleCursorY++;

  //check (and correct) position in console and adjust so that cursor is visible
  ConsoleCheckPosCursorVisible( );

  //console has changed
  ConsoleChanged = 1;
}

//refresh console
void ConsoleRefresh( )
{
  //local variables
  long X, Y, FntX, FntY, TexX, TexY, ChrIndex;
  char Chr;
  unsigned char LineData;

  //do nothing if refresh not neccessary
  if( ! ConsoleChanged )
    return;

  //check (and correct) position in console
  ConsoleCheckPos( );

  //put characters in console-buffer
  for( Y = 0; Y < ConsoleTextShowSizeY; Y++ )
  for( X = 0; X < ConsoleTextShowSizeX; X++ )
  {
    //character changed
    if( ConsoleTextShow[Y][X] != ConsoleText[Y + ConsoleTopLeftY][X + ConsoleTopLeftX] )
    {
      //get new character
      Chr = ConsoleText[Y + ConsoleTopLeftY][X + ConsoleTopLeftX];
      //put character into text-show-buffer
      ConsoleTextShow[Y][X] = Chr;
      //put character into console-buffer
      ChrIndex = (unsigned char)Chr;
      for( FntY = 0; FntY < FontSizeY; FntY++ )
      {
        LineData = FontData[ChrIndex][FntY];
        TexY = ConsoleInnerSizeY - (Y * FontSizeY + FntY) + ConsoleBorder;
        for( FntX = 0; FntX < FontSizeX; FntX++ )
        {
          TexX = X * FontSizeX + FntX + ConsoleBorder;
          if( (LineData & ((1 << (FontSizeX - 1)) >> FntX)) != 0 )
          {
            ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleTextB;
            ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleTextG;
            ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleTextR;
          }
          else
          {
            ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleBackB;
            ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleBackG;
            ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleBackR;
          }
        }
      }
    }
  }

  //remove old cursor from console-buffer
  if( ConsoleCursorShowX >= 0 && ConsoleCursorShowX < ConsoleTextShowSizeX &&
    ConsoleCursorShowY >= 0 && ConsoleCursorShowY < ConsoleTextShowSizeY )
  {
    //get character
    ChrIndex = (unsigned char)ConsoleTextShow[ConsoleCursorShowY][ConsoleCursorShowX];
    //put character into console-buffer
    for( FntY = 0; FntY < FontSizeY; FntY++ )
    {
      LineData = FontData[ChrIndex][FntY];
      TexY = ConsoleInnerSizeY - (ConsoleCursorShowY * FontSizeY + FntY) + ConsoleBorder;
      for( FntX = 0; FntX < FontSizeX; FntX++ )
      {
        TexX = ConsoleCursorShowX * FontSizeX + FntX + ConsoleBorder;
        if( (LineData & ((1 << (FontSizeX - 1)) >> FntX)) != 0 )
        {
          ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleTextB;
          ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleTextG;
          ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleTextR;
        }
        else
        {
          ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleBackB;
          ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleBackG;
          ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleBackR;
        }
      }
    }
  }

  //put cursor into console-buffer
  ConsoleCursorShowX = ConsoleCursorX - ConsoleTopLeftX;
  ConsoleCursorShowY = ConsoleCursorY - ConsoleTopLeftY;
  if( ConsoleCursorShowX >= 0 && ConsoleCursorShowX < ConsoleTextShowSizeX &&
    ConsoleCursorShowY >= 0 && ConsoleCursorShowY < ConsoleTextShowSizeY )
  {
    for( FntY = 0; FntY < FontSizeY; FntY++ )
    {
      LineData = CursorData[FntY];
      TexY = ConsoleInnerSizeY - (ConsoleCursorShowY * FontSizeY + FntY) + ConsoleBorder;
      for( FntX = 0; FntX < FontSizeX; FntX++ )
      {
        TexX = ConsoleCursorShowX * FontSizeX + FntX + ConsoleBorder;
        if( (LineData & ((1 << (FontSizeX - 1)) >> FntX)) != 0 )
        {
          ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleCursorB;
          ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleCursorG;
          ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleCursorR;
        }
        else
        {
          ConsolePixelComponent( TexY, TexX, 0 ) = ConsoleBackB;
          ConsolePixelComponent( TexY, TexX, 1 ) = ConsoleBackG;
          ConsolePixelComponent( TexY, TexX, 2 ) = ConsoleBackR;
        }
      }
    }
  }

  //load console-buffer into tex1
  LoadTexture( ConsoleData, ConsoleTexSizeX, ConsoleTexSizeY, ConsoleTexNo );

  //console is refreshed
  ConsoleChanged = 0;
}

//refresh console
void GetConsoleLine( char * pLine, long LineSize )
{
  //local variables
  long I;

  //check (and correct) position in console
  ConsoleCheckPos( );

  //get number of characters in last line
  I = strlen( ConsoleText[ConsoleCursorY] );
  
  //do not exceed maximum-size
  if( I > LineSize - 1 )
    I = LineSize - 1;

  //copy last line
  memcpy( pLine, ConsoleText[ConsoleCursorY], I );
  pLine[I] = 0;
}

//move console left
void ConsoleMoveLeft( )
{
  ConsoleTopLeftX -= ConsoleLeftRightStep;
  if( ConsoleTopLeftX < 0 )
    ConsoleTopLeftX = 0;
  ConsoleChanged = 1;
}

//move console right
void ConsoleMoveRight( )
{
  ConsoleTopLeftX += ConsoleLeftRightStep;
  if( ConsoleTopLeftX > ConsoleTextSizeX - ConsoleTextShowSizeX )
    ConsoleTopLeftX = ConsoleTextSizeX - ConsoleTextShowSizeX;
  ConsoleChanged = 1;
}

//move console up
void ConsoleMoveUp( )
{
  ConsoleTopLeftY -= ConsoleUpDownStep;
  if( ConsoleTopLeftY < 0 )
    ConsoleTopLeftY = 0;
  ConsoleChanged = 1;
}

//move console down
void ConsoleMoveDown( )
{
  ConsoleTopLeftY += ConsoleUpDownStep;
  if( ConsoleTopLeftY > ConsoleTextSizeY - ConsoleTextShowSizeY )
    ConsoleTopLeftY = ConsoleTextSizeY - ConsoleTextShowSizeY;
  ConsoleChanged = 1;
}

//move console up a page
void ConsoleMovePageUp( )
{
  ConsoleTopLeftY -= ConsolePageStep;
  if( ConsoleTopLeftY < 0 )
    ConsoleTopLeftY = 0;
  ConsoleChanged = 1;
}

//move console down a page
void ConsoleMovePageDown( )
{
  ConsoleTopLeftY += ConsolePageStep;
  if( ConsoleTopLeftY > ConsoleTextSizeY - ConsoleTextShowSizeY )
    ConsoleTopLeftY = ConsoleTextSizeY - ConsoleTextShowSizeY;
  ConsoleChanged = 1;
}

