/* TroiaPreview
 * version 0.4 date 2004-09-20
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>

#include "Tools.h"
#include "Globals.h"
#include "Pixel.h"
#include "Frame.h"
#include "Life.h"

//private global variables
double LifeInterval = 0.0; //GameOfLife stopped
double LifeLastStepTime; //time of last automatic step

//callback for step 1: use D7 as source and D6 as destination for calculation
unsigned char LifeCallbackCalc( tPixel3D Pixel, void * pUnused )
{
  unsigned char Neighbours[] = { 1, 2, 3, 4, 6, 7, 8, 9 };
  unsigned char Cnt, Dir, Color;
  tPixel3D Pix;
  int I;

  //keep compiler happy
  pUnused = NULL;

  //count pixels in neighbourhood (get source from D7)
  Cnt = 0;
  for( I = 0; I < (int)count( Neighbours ); I++ )
  {
    Pixel3DMove( Pixel, Neighbours[I], &Pix, &Dir );
    if( FrameGetPixel( Pix ) & 0x80 )
      Cnt++;
  }

  //use GameOfLife rules to get new pixel state (put it into D6)
  Color = FrameGetPixel( Pixel );
  if( Cnt == 3 ) //birth: 3 neighbours
    Color |= 0x40;
  else if( Cnt != 2 || !(Color & 0x80) ) //death: not 2 or 3 neighbours or already dead
    Color &= 0xBF;
  return Color;
}

//callback for step 2: copy D0 into D7
unsigned char LifeCallbackCopy( tPixel3D Pixel, void * pUnused )
{
  unsigned char Color;

  //keep compiler happy
  pUnused = NULL;

  Color = FrameGetPixel( Pixel );
  Color = (Color & 0x7F) | (Color & 0x40) << 1;
  return Color;
}

//do a GameOfLife step with the current frame
void LifeStep( )
{
  //there is only one buffer, but there are greyscales
  //so we will use 2 steps
  //1: use D7 as source and D6 as destination for calculation
  //2: copy D6 into D7

  FrameColorByCallback( LifeCallbackCalc, NULL );
  FrameColorByCallback( LifeCallbackCopy, NULL );
}

//start automatic GameOfLife
void LifeStart( double Interval )
{
  LifeInterval = DoubleMinMax( Interval, LifeMinInterval, LifeMaxInterval );
  LifeLastStepTime = TimeGet( ) - LifeInterval;
}

//stop automatic GameOfLife
void LifeStop( )
{
  LifeInterval = 0.0;
}

//called periodically to do some work
void LifeTask( )
{
  double CurrentTime;

  //exit if GameOfLife not running
  if( LifeInterval < LifeMinInterval * 0.5 ) //use * 0.5 to get a safe decision threshold
    return;

  //get time
  CurrentTime = TimeGet( );

  //exit if interval not elapsed
  if( CurrentTime < LifeInterval + LifeLastStepTime )
    return;

  //do a step
  LifeStep( );

  //restart interval
  LifeLastStepTime = CurrentTime;
}
